#!/usr/bin/perl
#
# Minimalist - MINImalist MAiling LIST manager.
# Copyright (c) 1999 Vladimir Litovka <doka@webest.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 675
# Mass Ave, Cambridge, MA 02139, USA.

$version = "1.3.5-pl1";			# Modified by C.Strozzi, 19 May 1999.
$config = "/usr/local/etc/minimalist.conf";

# Lists' status bits
$OPEN = 0;
$RO = 1;
$CLOSED = 2;
$MANDATORY = 4;

#####################################################
# Default values
#
$sendmail = '/usr/sbin/sendmail';
$domain = `hostname --fqdn`; chomp $domain;
$directory = '/var/spool/minimalist';
$admin = "root\@$domain";
$listpwd = $userpwd = '';
$security = 'careful';
$archive = 'no';
$archpgm = 'BUILTIN';
$status = $OPEN;
$copy_sender = 'yes';
$reply_to_list = 'no';
$modify_subject = 'yes';
$maxusers = 0;
$maxrcpts = 20;

#####################################################
# Various regular expressions
#
# for matching rounding spaces

$spaces = '^\s*(.+?)\s*$';

# for parsing two forms of mailing addresses:
#
# 1st form: Vladimir Litovka <doka@webest.com>
# 2nd form: doka@webest.com (Vladimir Litovka)

$first = '.*<\s*([-\.\w\@]*)\s*>.*';
$second = '^\s*([-\.\w\@]+)\s*\(.*\).*';

# Help
$helpmsg = <<_EOF_ ;
This is the Minimalist Mailing List Manager.

All commands must appear in the SUBJECT of mail messages.

Supported commands are:

subscribe <list> [<email>] :
    subscribe user to <list>

unsubscribe <list> [<email>] :
    unsubscribe user from <list>

info [<list>] :
    gives you information about existing lists or about <list>

list <list> :
    gives you the list of users subscribed to <list>

help :
    This message

Note, that <email> can be used only by trusted users. Otherwise, the
address in the 'From:' field will be used. The LIST command can be
used only by trusted users. Trusted users must supply the list password
in the first line of the message body, i.e:

*password: list_password

followed by any number of empty rows. The password will be removed from the
message, of course.

-- 
Sincerely, the Minimalist
_EOF_

########################################################
# >>>>>>>>>>>>>>> SELF - CONFIGURING <<<<<<<<<<<<<<<<< #
########################################################

read_config($config, "global");

# Admin always in trusted users
$trusted[@trusted] = $admin;

$me = "minimalist\@$domain";

#######################################################
# >>>>>>>>>>>>>>>> CHECK CONFIGURATION <<<<<<<<<<<<<< #
#######################################################

if ($ARGV[0] eq '-') {
  print "\nMinimalist v$version, pleased to meet you.\n\n".
	"================= Global configuration ================\n".
	"Directory: $directory\n".
	"Admin: $admin\n";
  die "$directory doesn't exist!\n" if (! -d $directory);

  while ( $ARGV[0] ) {
    $list = $ARGV[0];
    if ($list ne '-') {
      if (!chdir("$directory/$list")) {
        print "\nThere isn't such list \U$list\E\n";
        shift; next;
       }
      read_config("config");
      print "================= \U$list\E ================\n";
     }
    print "Sendmail: $sendmail\n".
	"Domain: $domain\n".
	"Password: ".($listpwd ? "yes, list protected" : "none")."\n".
	"Security: $security\n".
	"Archiving: $archive\n".
	  ($archive ne 'no' ? " * Archiver: $archpgm\n" : "").
	"Status:";
    if ($status) {
      print " read-only" if ($status & $RO);
      print " closed" if ($status & $CLOSED);
      print " mandatory" if ($status & $MANDATORY);
     }
    else { print " open"; }
    print "\nCopy to sender: $copy_sender\n".
	"Reply-To list: $reply_to_list\n".
	"Modify subject: $modify_subject\n".
	"Maximal users per list: ".($maxusers ? $maxusers : "unlimited")."\n".
	"Maximal recipients per message: ".($maxrcpts ? $maxrcpts : "unlimited")."\n".
	"Trusted users:\n";
    foreach (@trusted) { print "\t + ".$_."\n"; }
    print "\n";
    
    # Various checks
    die "$sendmail doesn't exist!\n" if (! -x $sendmail);
    die "Illegal domain '$domain'\n" if ($domain !~ /^(\w[-\w]*\.)+[a-z]{2,4}$/i);
    die "Illegal security level '$security'\n" if ($security !~ /^none$|^careful$/i);
    die "Illegal archiving strategy '$archive'\n" if ($archive !~ /^no$|^daily$|^monthly$|^yearly$|^pipe$/i);
    die "Illegal 'copy to sender' value '$copy_sender'\n" if ($copy_sender !~ /^yes$|^no$/i);
    die "Illegal 'modify subject' value '$modify_subject'\n" if ($modify_subject !~ /^yes$|^no$|^more$/i);
    die "Illegal 'reply-to list' value '$reply_to_list'\n" if ($reply_to_list !~ /^yes$|^no$|\@/i);

    shift;
   }

  exit 0;
 }

########################################################
# >>>>>>>>>>>>>>>>>> START HERE <<<<<<<<<<<<<<<<<<<<<< #
########################################################

$list = $ARGV[0];

while (<STDIN>) { $message .= $_ }

($header, $body) = split(/\n\n/, $message, 2); $header .= "\n";
@headers = split (/\n/, $header);

# Get password if it's there and strip this line from message's body
if ($body =~ s/(^[ \t]*\*password: )(.*)\n+(.*)/$3/i) {
  $userpwd = $2;
 }

$body =~ s/\n*$//g;

#####################################################
# Mail FROM, but REPLY-TO

$from = $reply = $sender = $xsender = $subject = '';

foreach (@headers){
  if    ( ($from eq '') && /^from:[ \t]*(.*)/i ) { $from = $1; }
  elsif ( ($reply eq '') && /^reply-to:[ \t]*(.*)/i ) { $reply = $1; }
  elsif ( /^sender:[ \t]*(.*)/i ) { $sender = $1; }
  elsif ( /^x-sender:[ \t]*(.*)/i ) { $xsender = $1; }
 }

exit 0 if (($xsender eq $me) || ($from eq $me));	# LOOP detected
exit 0 if (($from =~ /MAILER-DAEMON/i) ||
	   ($sender =~ /MAILER-DAEMON/i) ||
	   ($xsender =~ /MAILER-DAEMON/i));		# ignore messages from MAILER-DAEMON

($reply eq '') ? ($mailto = $from) : ($mailto = $reply);

# Preparing From:
if ($from =~ /<.*>/) { $from =~ s/$first/$1/og; }
elsif ($from =~ /\(.*\)/) { $from =~ s/$second/$1/og; }
$from =~ s/$spaces/$1/g;
$from = lc($from);

# Get (multiline) subject
if ($header =~ /subject: (.*\n([ \t]+.*\n)*)/i) {
  $subject = $1; chomp $subject; }

#####################################################
# Message to list
#
if ($list ne '') {

 chdir("$directory/$list");
 read_config("config");

 open LIST, "list";
 $users .= $_ while (<LIST>);
 close LIST;
 @members = split ("\n", $users);
 
 $subject =~ s/^[ \t]*//g;
 if ($modify_subject ne 'no') {
   if ($modify_subject eq 'more') { $subject =~ s/^Re:[ \t]*//ig }
   if ($subject !~ /^(Re: )?\[$list\] /i) {
     $subject = "[$list] ".$subject;
    }
  }
  
 if (($security eq 'careful') && !grep(/^$from$/i, @members)) {
   $msg = <<_EOF_ ;
From: $me
To: $mailto
Subject: $subject

ERROR:
    You ($from) are not subscribed to this list.

SOLUTION:
    Send a message to minimalist\@$domain with a subject
    of 'info' (no quotes) for information about subscribing.

===========================================================================

$body
_EOF_
  } 
 elsif (($status & $RO) && (!grep(/^$from$/i, @trusted) || ($userpwd ne $listpwd))) {
   $msg = <<_EOF_ ;
From: $me
To: $mailto
Subject: $subject

ERROR:
    You are not allowed to write to this list.

===========================================================================

$body
_EOF_
  }
 else {

   # OK, all checks done. Now send message to members except those who already
   # received this message directly from sender

   # Get all recipients from To: and Cc: (expect multi-line)
   $header =~ /((.*\n)*)to: (.*\n([ \t]+.*\n)*)/i; $rc = $3;
     @recip = split(/[ \n\t,]+/, $rc);
   $header =~ /((.*\n)*)cc: (.*\n([ \t]+.*\n)*)/i; $rc = $3;
     push (@recip, split(/[ \n\t,]+/, $rc));

   foreach (@recip) {
     if ( /<.*>/ ) { s/$first/$1/g; }
     elsif ( /\(.*\)/ ) { s/$second/$1/g; }
     s/$spaces/$1/g;
     push (@rcpts, lc($_));
    }

   if ($copy_sender eq 'no') { push (@rcpts, $from) }

   @members = sort @members; @rcpts = sort @rcpts;

   for ($r=0, $m=0; $m < @members; ) {
     if ($members[$m] lt $rcpts[$r] || $r >= @rcpts) {
       push (@recipients, $members[$m]);
       $m++; }
     elsif ($members[$m] eq $rcpts[$r]) { $r++; $m++; }
     elsif ($members[$m] gt $rcpts[$r]) { $r++ };
    }

   # Archive message if desired
   if ($archive eq 'pipe') { arch_pipe(); }
   elsif ($archive ne 'no') { archive(); }

   # Remove conflicting X- headers
   $header =~ s/x-mailing-list.*\n([ \t]+.*\n)*//ig;
   $header =~ s/x-sender: .*\n([ \t]+.*\n)*//ig;

   # Remove all recipients - they already received message
   $header =~ s/((.*\n)*?)(to:|cc:) (.*\n([ \t]+.*\n)*)/$1/ig;

   $header =~ s/\n*$//g;
   $header .= "\nTo: $list\@$domain\n";		# Mail to himself, loop is handling

   if ($reply_to_list ne 'no') {
     $header =~ s/((.*\n)*?)reply-to: .*\n([ \t]+.*\n)*/$1/ig;
     if ($reply_to_list eq 'yes') { $header .= "Reply-To: $list\@$domain\n"; }
     else { $header .= "Reply-To: $reply_to_list\n"; }
    }

   if ($modify_subject ne 'no') {
     $header =~ s/((.*\n)*?)subject: .*\n([ \t]+.*\n)*/$1/ig;
     $header .= "Subject: $subject\n";
    }

   #########################################################
   # Send message to recipients ($maxrcpts per message)

   $envelope_sender = "-f $me";
   $footer = read_info("footer");
   $maxrcpts = $#recipients + 1 if ($maxrcpts == 0);	# Unlimited?

   foreach $one (@recipients) {
     if ($rcs == $maxrcpts) {
       sendPortion();
       $bcc = ''; $rcs = 0;	# Clear counters
      }
     if ($one ne '') {
       $bcc .= "$one,"; $rcs++; }
    }

   sendPortion() if ($bcc ne '');	# Send to rest subscribers

   $msg = '';	# Clear message, don't send anything anymore
  }

} else {

#####################################################
# Message to minimalist
#
# Allowed commands:
#	subscribe <list> [<e-mail>]
#	unsubscribe <list> [<e-mail>]
#	info [<list>]
#	list <list>
#	help

 $list = ''; $email = '';
 ($cmd, $list, $email) = split (/ /, $subject, 3);
 $cmd =~ s/$spaces/$1/g; $cmd = lc($cmd);
 $list =~ s/$spaces/$1/g; $list = lc($list);

 if ($email ne '') {
   if ($email =~ /<.*>/) { $email =~ s/$first/$1/og; }
   elsif ($email =~ /\(.*\)/) { $email =~ s/$second/$1/og; }
   $email =~ s/$spaces/$1/g;
   $email = lc($email);
  }

 $msg = "From: $me\n".
	"To: $mailto\n".
	"Subject: Re: $subject\n";
 
 if ($cmd eq 'help') {
   $msg .= "\n".$helpmsg; }

 elsif ( ($list ne '') && (! -d "$directory/$list") ) {
   $msg .= <<_EOF_ ;

ERROR:
    There is no such list \U$list\E here.

SOLUTION:
    Send a message to minimalist\@$domain with a subject
    of 'info' (no quotes) for a list of available mailing lists.
_EOF_
  }

 elsif ( (($cmd eq 'subscribe') || ($cmd eq 'unsubscribe')) && ($list ne '') ) {
   $melist = "$list\@$domain";
   if (($from ne $me) && ($from ne $melist) && ($email ne $me) && ($email ne $melist)) {
     chdir "$directory/$list";
     read_config("config");
     $owner = "$list-owner\@$domain";

     if (grep(/^$from$/i, @trusted) && ($userpwd eq $listpwd)) { eval("$cmd(1)"); }
     elsif (($email ne '') && ($email ne $from)) {
       $msg .=  "\nERROR:\n\tYou aren't allowed to subscribe other persons.\n".
       		"\nSOLUTION:\n\tNone.\n";
      }
     elsif (($cmd eq 'subscribe') && ($status & $CLOSED)) {
       $msg .=  "\nERROR:\n\tSorry, this list is closed for you.\n".
       		"\nSOLUTION:\n\tAre you unsure? Please, complain to $owner\n";
      }
     elsif (($cmd eq 'unsubscribe') && ($status & $MANDATORY)) {
       $msg .=  "\nERROR:\n\tSorry, this list is mandatory for you.\n".
       		"\nSOLUTION:\n\tAre you unsure? Please, complain to $owner\n";
      }
     else {
       eval("$cmd(0)");
      }
    }	# loop preventing
  }	# subscribe/unsubscribe

 elsif ($cmd eq 'info') {
   if ($list ne '') {
     $msg .= "\nHere is the available information about \U$list\E\n\n";
     $msg .= read_info("$directory/$list/info");
    }
   else {
     $msg .= "\nThese are the mailing lists available at $domain:\n\n";
     if (open(INFO, "$directory/lists.lst")) {
       while (<INFO>) {
         $msg .= $_ if (! /^#/); }
       close INFO;
      }
    }
  }
 elsif (($cmd eq 'list') && ($list ne '')) {
   chdir "$directory/$list";
   read_config("config");

   if (grep(/^$from$/i, @trusted) && ($userpwd eq $listpwd)) {
     $msg .= "\nUsers, subscribed to \U$list\E:\n\n";
     if (open(LIST, "list")) {
       $msg .= $_ while (<LIST>);
       close LIST;
      }
    }
   else {
     $msg .= "\nERROR:\n\tYou aren't allowed to get listing of subscribed users.\n".
     	     "\nSOLUTION:\n\tNone.\n";
    }
  }
 else {
   $msg .= "\nERROR:\n\tBad syntax or unknown instruction.\n".
   	   "\nSOLUTION:\n\n$helpmsg\n";
  }
}

if ($msg ne '') {
  open MAIL, "| $sendmail -t";
  print MAIL $msg."\n";
  close MAIL;
 }

exit 0;

#########################################################################
# Supplementary functions
#

#################### SUBSCRIBE #####################
sub subscribe {

 my ($trusteduser) = @_;
 my ($cc);

 if ($email ne '') {
   $cc = "$email,";
   $from = $email;
  }

 if (open LIST, "list") {
   $users .= $_ while (<LIST>);
   close LIST;
   @members = split ("\n", $users);
   if (grep(/^$from$/i, @members)) {
     $deny = 1;
     $cause = "you already subscribed to \U$list\E";
    }
   elsif ((! $trusteduser) && ($maxusers > 0) && (@members >= $maxusers)) {
     $deny = 1; $cc .= "$owner,";
     $cause = "there are already maximal count of subscribers ($maxusers) at \U$list\E";
    }
   else { open LIST, ">>list"; }
  }
 else { open LIST, ">list"; } 

 if ($cc) {
   chop $cc;
   $msg .= "Cc: $cc\n";
  }

 $msg .= "\nDear $from,\n\n";

 if (! $deny) {
   print LIST "$from\n";
   close LIST;
   $msg .= <<_EOF_ ;
    You have subscribed to \U$list\E successfully.

Please, note the following:
_EOF_
   $msg .= "\n".read_info("info");
  }
 else {
   $msg .= <<_EOF_ ;
You haven't subscribed to \U$list\E due to the following reason:
    * $cause.

If you have comments or questions, please, send them to the list
administrator $owner.
_EOF_
  }
}

#################### UNSUBSCRIBE #####################
sub unsubscribe {
 if ($email ne '') {
   $msg .= "Cc: $email\n" ;
   $from = $email;
  }

 if (open LIST, "list") {
   $users .= $_ while (<LIST>);
   close LIST;

   $users =~ s/$from\n//g;
   open LIST, ">list";
   print LIST $users;
   close LIST;
  }

 $msg .= "\nUser $from has successfully unsubscribed.\n\nGood bye!";
}

######################### READ_CONFIG ############################
sub read_config {

my ($fname, $global) = @_;

if (open(CONF, $fname)) {
  while (<CONF>) {
    if($_ =~ /^#/) {
    } elsif (($_ =~ /^directory/i) && $global) {
     ($directive, $directory) = split(/=/, $_);
     $directory =~ s/$spaces/$1/g;

    } elsif (($_ =~ /^admin/i) && $global) {
     ($directive, $admin) = split(/=/, $_);
     $admin =~ s/$spaces/$1/g;

    } elsif ($_ =~ /^sendmail/i) {
     ($directive, $sendmail) = split(/=/, $_);
     $sendmail =~ s/$spaces/$1/g;

    } elsif ($_ =~ /^domain/i) {
     ($directive, $domain) = split(/=/, $_);
     $domain =~ s/$spaces/$1/g;

    } elsif ($_ =~ /^password/i) {
     ($directive, $listpwd) = split(/=/, $_);
     $listpwd =~ s/$spaces/$1/g;

    } elsif ($_ =~ /^security/i) {
     ($directive, $security) = split(/=/, $_);
     $security =~ s/$spaces/$1/g;
     $security = lc($security);

    } elsif ($_ =~ /^archive/i) {
     ($directive, $archive) = split(/=/, $_);
     $archive =~ s/$spaces/$1/g;

     if ($archive =~ s/^pipe\s+//i) {
       $archpgm = $archive;
       $archive = "pipe"; }
     else {
       $archpgm = 'BUILTIN';
       $archive =~ s/$spaces/$1/g;
       $archive = lc($archive); }

    } elsif ($_ =~ /^maxusers/i) {
     ($directive, $maxusers) = split(/=/, $_);
     chomp $maxusers; $maxusers =~ s/$spaces/$1/g;

    } elsif ($_ =~ /^maxrcpts/i) {
     ($directive, $maxrcpts) = split(/=/, $_);
     chomp $maxrcpts; $maxrcpts =~ s/$spaces/$1/g;

    } elsif ($_ =~ /^status/i) {
     ($directive, $status) = split(/=/, $_);
     chomp $status; $status =~ s/$spaces/$1/g;
     $status = lc($status);

     # Calculate mask for status
     %strel = ("open", $OPEN, "ro", $RO, "closed", $CLOSED, "mandatory", $MANDATORY);
     @starr = split(/,/, $status);
     $status = 0;
     foreach (@starr) { $status += $strel{$_}; }

    } elsif ($_ =~ /^copy to sender/i) {
     ($directive, $copy_sender) = split(/=/, $_);
     chomp $copy_sender; $copy_sender =~ s/$spaces/$1/g;
     $copy_sender = lc($copy_sender);

    } elsif ($_ =~ /^reply-to list/i) {
     ($directive, $reply_to_list) = split(/=/, $_);
     chomp $reply_to_list; $reply_to_list =~ s/$spaces/$1/g;
     $reply_to_list = lc($reply_to_list);
     # In global config only 'yes' or 'no' allowed
     if ($global && ($reply_to_list ne 'yes')) { $reply_to_list = 'no'; }

    } elsif ($_ =~ /^modify subject/i) {
     ($directive, $modify_subject) = split(/=/, $_);
     chomp $modify_subject; $modify_subject =~ s/$spaces/$1/g;
     $modify_subject = lc($modify_subject);

    } elsif ($_ =~ /^trusted/i) {
     ($directive, $trust) = split(/=/, $_, 2);
     chomp $trust; $trust =~ s/[ \t]*//g;
     @junk = split(':', $trust);
     foreach $s (@junk) {
       if ( $s =~ s/^\@// ) {	# Expand items, starting with '@'
	 if (open(IN, $s)) {
	   while (<IN>) {
	     chomp $_; $trusted[@trusted] = $_; }
	   close IN;
	  }
	}
       elsif ($s ne '') { $trusted[@trusted] = $s; }
      }
    }
  }
  close CONF;
 }
}

####### Read file and substitute all macroses #########
sub read_info {
 my ($fname) = @_;
 my ($tail);

 if (open(TAIL, $fname)) {
   $tail .= $_ while (<TAIL>);
   close TAIL;

   $tail =~ s/\\a/$admin/ig;
   $tail =~ s/\\d/$domain/ig;
   $tail =~ s/\\l/$list/ig;
   $tail =~ s/\\o/$list-owner\@$domain/ig;
  }

 $tail;
}

###### Send ready portion of message ######
sub sendPortion {

 chop $bcc;
 $temphdr = $header . "Bcc: $bcc\n";
 $temphdr .= "X-Sender: $me\n".
 "X-Mailing-List: $list\@$domain\n".
 "X-Mailing-List-Owner: $list-owner\@$domain\n".
 "X-Mailing-List-Server: Minimalist v$version\n\n";
 $msg = $temphdr.$body."\n\n".$footer;

 open MAIL, "| $sendmail -t $envelope_sender";
 print MAIL $msg."\n";
 close MAIL;
}

##################### Built-in archiver ##########################
sub archive {

 @date = localtime;
 $year = 1900 + $date[5];
 $month = 1 + $date[4];
 $day = $date[3];

 $path = "archive/";
 mkdir($path, 0755) if (! -d $path);

 @types = ("yearly", "monthly", "daily");
 %rel = ($types[0], $year, $types[1], $month, $types[2], $day);

 foreach $key (@types) {
   $path .= $rel{$key}."./";
   mkdir($path, 0755) if (! -d $path);
   last if ($key eq $archive);
  }

 if (open(NUM, $path."SEQUENCE")) {
   read NUM, $msgnum, 16;
   $msgnum = int($msgnum);
   close NUM;
  }
 else { $msgnum = 0 }

 open ARCHIVE, ">$path".$msgnum;
 print ARCHIVE $header."\n".$body;
 close ARCHIVE;

 open NUM, ">$path"."SEQUENCE";
 print NUM $msgnum+1;
 close NUM;
}


##################### External archiver ##########################
sub arch_pipe {

 open( ARCHIVE, "| $archpgm" );
 print ARCHIVE $header."\n".$body;
 close( ARCHIVE );
}
